@@ -1,5 +1,5 @@
 #ifndef lint
-static char *RCSid() { return RCSid("$Id: graph3d.c,v 1.157.2.12 2007/08/30 19:24:02 sfeam Exp $"); }
+static char *RCSid() { return RCSid("$Id: graph3d.c,v 1.157.2.20 2008/09/10 16:01:04 sfeam Exp $"); }
 #endif
 
 /* GNUPLOT - graph3d.c */
@@ -339,13 +339,18 @@ boundary3d(struct surface_points *plots, int count)
 	max_ptitl_len = ytlen;
     key_col_wth = (max_ptitl_len + 4) * t->h_char + key_sample_width;
 
-    /* luecken@udel.edu modifications
-       sizes the plot to take up more of available resolution */
-    if (lmargin >= 0)
-	plot_bounds.xleft = t->h_char * lmargin;
+    if (lmargin.scalex == screen)
+	plot_bounds.xleft = lmargin.x * (float)t->xmax + 0.5;
+    else if (lmargin.x >= 0)
+	plot_bounds.xleft = lmargin.x * (float)t->h_char + 0.5;
     else
 	plot_bounds.xleft = t->h_char * 2 + t->h_tic;
-    plot_bounds.xright = xsize * t->xmax - t->h_char * 2 - t->h_tic;
+
+    if (rmargin.scalex == screen)
+	plot_bounds.xright = rmargin.x * (float)t->xmax + 0.5;
+    else /* No tic label on the right side, so ignore rmargin */
+	plot_bounds.xright = xsize * t->xmax - t->h_char * 2 - t->h_tic;
+
     key_rows = ptitl_cnt;
     key_cols = 1;
     if ((key->region == GPKEY_AUTO_EXTERIOR_MARGIN || key->region == GPKEY_AUTO_EXTERIOR_LRTBC)
@@ -369,8 +374,10 @@ boundary3d(struct surface_points *plots, int count)
     /* this should also consider the view and number of lines in
      * xformat || yformat || xlabel || ylabel */
 
-    if (splot_map && bmargin >= 0)
-	plot_bounds.ybot = t->v_char * bmargin;
+    if (bmargin.scalex == screen)
+	plot_bounds.ybot = bmargin.x * (float)t->ymax + 0.5;
+    else if (splot_map && bmargin.x >= 0)
+	plot_bounds.ybot = (float)t->v_char * bmargin.x;
     else
 	plot_bounds.ybot = t->v_char * 2.5 + 1;
 
@@ -385,7 +392,12 @@ boundary3d(struct surface_points *plots, int count)
 		titlelin++;
 	}
     }
-    plot_bounds.ytop = ysize * t->ymax - t->v_char * (titlelin + 1.5) - 1;
+
+    if (tmargin.scalex == screen)
+	plot_bounds.ytop = tmargin.x * (float)t->ymax + 0.5;
+    else /* FIXME: Why no provision for tmargin in terms of character height? */
+	plot_bounds.ytop = ysize * t->ymax - t->v_char * (titlelin + 1.5) - 1;
+
     if (key->region == GPKEY_AUTO_INTERIOR_LRTBC
 	|| ((key->region == GPKEY_AUTO_EXTERIOR_LRTBC || key->region == GPKEY_AUTO_EXTERIOR_MARGIN)
 	    && key->margin == GPKEY_RMARGIN)) {
@@ -411,12 +423,21 @@ boundary3d(struct surface_points *plots, int count)
     plot_bounds.ybot += t->ymax * yoffset;
     xmiddle = (plot_bounds.xright + plot_bounds.xleft) / 2;
     ymiddle = (plot_bounds.ytop + plot_bounds.ybot) / 2;
+
+    
     /* HBB 980308: sigh... another 16bit glitch: on term's with more than
      * 8000 pixels in either direction, these calculations produce garbage
      * results if done in (16bit) ints */
-    xscaler = ((plot_bounds.xright - plot_bounds.xleft) * 4L) / 7L;	/* HBB: Magic number alert! */
+    /* HBB: Magic number alert! */
+    xscaler = ((plot_bounds.xright - plot_bounds.xleft) * 4L) / 7L;
     yscaler = ((plot_bounds.ytop - plot_bounds.ybot) * 4L) / 7L;
 
+    /* EAM Aug 2006 - Allow explicit control via set {}margin screen */
+    if (tmargin.scalex == screen || bmargin.scalex == screen)
+	yscaler = (plot_bounds.ytop - plot_bounds.ybot) / surface_scale;
+    if (rmargin.scalex == screen || lmargin.scalex == screen)
+	xscaler = (plot_bounds.xright - plot_bounds.xleft) / surface_scale;
+
     /* HBB 20011011: 'set size {square|ratio}' for splots */
     if (aspect_ratio != 0.0) {
 	double current_aspect_ratio;
@@ -489,6 +510,9 @@ place_labels3d(struct text_label *listhead, int layer)
 		     this_label->place.z, &xx, &yy);
 	    x = xx;
 	    y = yy;
+	    /* Only clip in 2D.   EAM - why? */
+	    if (splot_map && clip_point(x, y))
+		continue;
 	} else
 	    map3d_position(&this_label->place, &x, &y, "label");
 
@@ -500,9 +524,9 @@ static void
 place_arrows3d(int layer)
 {
     struct arrow_def *this_arrow;
+    BoundingBox *clip_save = clip_area;
 
-    /* Allow arrows to run off the plot, so long as they are still on
-     * the canvas */
+    /* Allow arrows to run off the plot, so long as they are still on the canvas */
     if (term->flags & TERM_CAN_CLIP)
       clip_area = NULL;
     else
@@ -522,6 +546,7 @@ place_arrows3d(int layer)
 	    FPRINTF((stderr,"place_arrows3d: skipping out-of-bounds arrow\n"));
 	}
     }
+    clip_area = clip_save;
 }
 
 /* we precalculate features of the key, to save lots of nested
@@ -627,6 +652,11 @@ do_3dplot(
     yscale3d = 2.0 / (Y_AXIS.max - Y_AXIS.min);
     xscale3d = 2.0 / (X_AXIS.max - X_AXIS.min);
 
+    /* Initialize palette */
+    if (!quick) {
+	can_pm3d = is_plot_with_palette() && !make_palette() && term->set_color;
+    }
+
     /* Give a chance for rectangles to be behind everything else */
     place_rectangles( first_object, -1, 3, NULL );
 
@@ -643,14 +673,6 @@ do_3dplot(
     else if (splot_map && border_layer == 0)
 	draw_3d_graphbox(plots, pcount, BORDERONLY, LAYER_BACK);
 
-    /* Draw PM3D color key box */
-    if (!quick) {
-	can_pm3d = is_plot_with_palette() && !make_palette()
-	    && term->set_color;
-	if (can_pm3d && is_plot_with_colorbox())
-	    draw_color_smooth_box(MODE_SPLOT);
-    }
-
 #ifdef USE_GRID_LAYERS
     if (!hidden3d && (grid_layer == -1))
 	/* Default layering mode.  Draw the back part now, but not if
@@ -736,6 +758,10 @@ do_3dplot(
 	}
     }
 
+    /* Add 'back' color box */
+    if (!quick && can_pm3d && is_plot_with_colorbox() && color_box.layer == LAYER_BACK)
+	draw_color_smooth_box(MODE_SPLOT);
+
     /* Add 'back' rectangles */
     place_rectangles(first_object, 0, 3, &clip_splot_map);
 
@@ -1178,7 +1204,7 @@ do_3dplot(
 			    set_color( cb2gray( z2cb(cntrs->z) ) );
 			else {
 			    if (prefer_line_styles && label_contours) {
-				struct lp_style_type ls;
+				struct lp_style_type ls = thiscontour_lp_properties;
 				lp_use_properties(&ls, ++thiscontour_lp_properties.l_type, FALSE);
 				term_apply_lp_properties(&ls);
 			    } else {
@@ -1260,6 +1286,7 @@ do_3dplot(
 		    case HISTEPS:
 			/* treat all the above like 'lines' */
 		    case LINES:
+		    case PM3DSURFACE:
 			cntr3d_lines(cntrs, &thiscontour_lp_properties);
 			break;
 
@@ -1283,8 +1310,6 @@ do_3dplot(
 		    case DOTS:
 			cntr3d_dots(cntrs);
 			break;
-		    case PM3DSURFACE: /* ignored */
-			break;
 #ifdef WITH_IMAGE
 		    case IMAGE:
 		    case RGBIMAGE:
@@ -1329,6 +1354,10 @@ do_3dplot(
 
 #endif /* USE_GRID_LAYERS */
 
+    /* Add 'front' color box */
+    if (!quick && can_pm3d && is_plot_with_colorbox() && color_box.layer == LAYER_FRONT)
+	draw_color_smooth_box(MODE_SPLOT);
+
     /* Add 'front' rectangles */
     place_rectangles(first_object, 1, 3, &clip_splot_map);
 
@@ -1857,10 +1886,10 @@ cntr3d_lines(struct gnuplot_contours *cntr, struct lp_style_type *lp)
 {
     int i;			/* point index */
     vertex this_vertex;
-    struct lp_style_type ls;
 
     /* user may prefer explicit line styles */
     if (prefer_line_styles && label_contours) {
+	struct lp_style_type ls = *lp;
 	lp_use_properties(&ls, lp->l_type, FALSE);
 	lp = &ls;
     }
@@ -2878,7 +2907,10 @@ map3d_getposition(
 	plot_coords = TRUE;
 	break;
     case graph:
-	*ypos = Y_AXIS.min + *ypos * (Y_AXIS.max - Y_AXIS.min);
+	if (splot_map)
+	    *ypos = Y_AXIS.max - *ypos * (Y_AXIS.max - Y_AXIS.min);
+	else
+	    *ypos = Y_AXIS.min + *ypos * (Y_AXIS.max - Y_AXIS.min);
 	plot_coords = TRUE;
 	break;
     case screen:
@@ -2972,9 +3004,9 @@ map3d_position_r(
 	else
 	    xpos = 0;
 	if (pos->scaley == graph)
-	    ypos = Y_AXIS.min;
+	    ypos = (splot_map) ? Y_AXIS.max : Y_AXIS.min;
 	else
-	    ypos = 0;
+	    ypos = (splot_map) ? Y_AXIS.max : 0;
 	if (pos->scalez == graph)
 	    zpos = Z_AXIS.min;
 	else
@@ -2984,8 +3016,8 @@ map3d_position_r(
 	*y -= yoriginlocal;
     } else {
     /* endpoint `screen' or 'character' coordinates */
-	    *x = xpos;
-	    *y = ypos;
+	*x = xpos;
+	*y = ypos;
     }
     return;
 }
