@@ -1,5 +1,5 @@
 #ifndef lint
-static char *RCSid() { return RCSid("$Id: show.c,v 1.185.2.2 2007/08/26 06:15:51 sfeam Exp $"); }
+static char *RCSid() { return RCSid("$Id: show.c,v 1.185.2.5 2008/06/01 06:13:45 sfeam Exp $"); }
 #endif
 
 /* GNUPLOT - show.c */
@@ -348,6 +348,10 @@ show_command()
     case S_COLORBOX:
 	show_colorbox();
 	break;
+    case S_COLORNAMES:
+	c_token--;
+	show_palette_colornames();
+	break;
     case S_POINTSIZE:
 	show_pointsize();
 	break;
@@ -1885,20 +1889,31 @@ show_margin()
 {
     SHOW_ALL_NL;
 
-    if (lmargin >= 0)
-	fprintf(stderr, "\tlmargin is set to %g\n", lmargin);
+    if (lmargin.scalex == screen)
+	fprintf(stderr, "\tlmargin is set to screen %g\n", lmargin.x);
+    else if (lmargin.x >= 0)
+	fprintf(stderr, "\tlmargin is set to %g\n", lmargin.x);
     else
 	fputs("\tlmargin is computed automatically\n", stderr);
-    if (bmargin >= 0)
-	fprintf(stderr, "\tbmargin is set to %g\n", bmargin);
+
+    if (bmargin.scalex == screen)
+	fprintf(stderr, "\tbmargin is set to screen %g\n", bmargin.x);
+    else if (bmargin.x >= 0)
+	fprintf(stderr, "\tbmargin is set to %g\n", bmargin.x);
     else
 	fputs("\tbmargin is computed automatically\n", stderr);
-    if (rmargin >= 0)
-	fprintf(stderr, "\trmargin is set to %g\n", rmargin);
+
+    if (rmargin.scalex == screen)
+	fprintf(stderr, "\trmargin is set to screen %g\n", rmargin.x);
+    else if (rmargin.x >= 0)
+	fprintf(stderr, "\trmargin is set to %g\n", rmargin.x);
     else
 	fputs("\trmargin is computed automatically\n", stderr);
-    if (tmargin >= 0)
-	fprintf(stderr, "\ttmargin is set to %g\n", tmargin);
+
+    if (tmargin.scalex == screen)
+	fprintf(stderr, "\ttmargin is set to screen %g\n", tmargin.x);
+    else if (tmargin.x >= 0)
+	fprintf(stderr, "\ttmargin is set to %g\n", tmargin.x);
     else
 	fputs("\ttmargin is computed automatically\n", stderr);
 }
@@ -2251,22 +2266,26 @@ show_colorbox()
     } else {
 	fputs("\tcolor box without border is ", stderr);
     }
+    if (color_box.where != SMCOLOR_BOX_NO) {
+	if (color_box.layer == LAYER_FRONT) fputs("drawn front\n\t", stderr);
+	else fputs("drawn back\n\t", stderr);
+    }
     switch (color_box.where) {
 	case SMCOLOR_BOX_NO:
 	    fputs("NOT drawn\n", stderr);
 	    break;
 	case SMCOLOR_BOX_DEFAULT:
-	    fputs("drawn at DEFAULT position\n", stderr);
+	    fputs("at DEFAULT position\n", stderr);
 	    break;
 	case SMCOLOR_BOX_USER:
-	    fputs("drawn at USER position: ", stderr);
+	    fputs("at USER position: ", stderr);
 	    show_position(&color_box.origin);
 	    fputs("\n\t          size: ", stderr);
 	    show_position(&color_box.size);
 	    fputs("\n", stderr);
 	    break;
 	default: /* should *never* happen */
-	    int_error(NO_CARET, "Argh !");
+	    int_error(NO_CARET, "Argh!");
     }
     fprintf(stderr,"\tcolor gradient is %s in the color box\n",
 	color_box.rotation == 'v' ? "VERTICAL" : "HORIZONTAL");
