@@ -1,5 +1,5 @@
 #ifndef lint
-static char *RCSid() { return RCSid("$Id: term.c,v 1.151.2.2 2007/03/22 05:05:07 sfeam Exp $"); }
+static char *RCSid() { return RCSid("$Id: term.c,v 1.151.2.5 2008/09/07 04:17:08 sfeam Exp $"); }
 #endif
 
 /* GNUPLOT - term.c */
@@ -1506,6 +1506,7 @@ change_term(const char *origname, int length)
 {
     int i;
     struct termentry *t = NULL;
+    TBOOLEAN ambiguous = FALSE;
 
     /* For backwards compatibility only */
     char *name = (char *)origname;
@@ -1517,12 +1518,17 @@ change_term(const char *origname, int length)
     for (i = 0; i < TERMCOUNT; i++) {
         if (!strncmp(name, term_tbl[i].name, length)) {
             if (t)
-                return (NULL);  /* ambiguous */
+                ambiguous = TRUE;
             t = term_tbl + i;
+	    /* Exact match is always accepted */
+	    if (length == strlen(term_tbl[i].name)) {
+		ambiguous = FALSE;
+		break;
+	    }
         }
     }
 
-    if (!t)                     /* unknown */
+    if (!t || ambiguous)
         return (NULL);
 
     /* Success: set terminal type now */
@@ -2444,7 +2450,7 @@ enhanced_recursion(
                             ;   /* do nothing */
                     }
                     localfontname = p;
-                    while ((ch = *p) > ' ' && ch != '=' && ch != '*')
+                    while ((ch = *p) > ' ' && ch != '=' && ch != '*' && ch != '}')
                         ++p;
                     save = *(savepos=p);
                     if (ch == '=') {
@@ -2593,19 +2599,19 @@ enhanced_recursion(
                 }
                 break;
             } else if (term->flags & TERM_IS_POSTSCRIPT) {
-                /* Shigeharu TAKENO  Aug 2004 - Needed in order for shift-JIS */
-                /* encoding to work. If this change causes problems then we   */
-                /* need a separate flag for shift-JIS and certain other 8-bit */
-                /* character sets.                                            */
-                /* EAM Nov 2004 - Nevertheless we must allow \ to act as an   */
-                /* escape for the 5 enhanced mode formatting characters even  */
-                /* though it corrupts certain Shift-JIS character sequences.  */
-                if (strchr("^_@&~",p[1]) == NULL) {
-                    (term->enhanced_open)(fontname, fontsize, base, widthflag, showflag, overprint);
-                    (term->enhanced_writec)('\\');
-                    (term->enhanced_writec)('\\');
-                    break;
-                }
+		/* Shigeharu TAKENO  Aug 2004 - Needed in order for shift-JIS */
+		/* encoding to work. If this change causes problems then we   */
+		/* need a separate flag for shift-JIS and certain other 8-bit */
+		/* character sets.                                            */
+		/* EAM Nov 2004 - Nevertheless we must allow \ to act as an   */
+		/* escape for a few enhanced mode formatting characters even  */
+		/* though it corrupts certain Shift-JIS character sequences.  */
+		if (strchr("^_@&~{}",p[1]) == NULL) {
+		    (term->enhanced_open)(fontname, fontsize, base, widthflag, showflag, overprint);
+		    (term->enhanced_writec)('\\');
+		    (term->enhanced_writec)('\\');
+		    break;
+		}
             }
             ++p;
 
