@@ -1,5 +1,5 @@
 #ifndef lint
-static char *RCSid() { return RCSid("$Id: unset.c,v 1.99.2.2 2007/06/07 16:27:38 sfeam Exp $"); }
+static char *RCSid() { return RCSid("$Id: unset.c,v 1.99.2.4 2008/03/12 03:09:17 sfeam Exp $"); }
 #endif
 
 /* GNUPLOT - unset.c */
@@ -102,7 +102,7 @@ static void unset_logscale __PROTO((void));
 static void unset_macros __PROTO((void));
 #endif
 static void unset_mapping __PROTO((void));
-static void unset_margin __PROTO((float *));
+static void unset_margin __PROTO((t_position *));
 static void unset_missing __PROTO((void));
 #ifdef USE_MOUSE
 static void unset_mouse __PROTO((void));
@@ -1076,9 +1076,10 @@ unset_mapping()
 
 /* process 'unset {blrt}margin' command */
 static void
-unset_margin(float *margin)
+unset_margin(t_position *margin)
 {
-    *margin = -1;
+    margin->scalex = character;
+    margin->x = -1;
 }
 
 /* process 'unset datafile' command */
@@ -1572,7 +1573,7 @@ unset_axislabel(AXIS_INDEX axis)
 {
     unset_axislabel_or_title(&axis_array[axis].label);
     axis_array[axis].label = default_axis_label;
-    if (axis == FIRST_Y_AXIS || axis == SECOND_Y_AXIS)
+    if (axis == FIRST_Y_AXIS || axis == SECOND_Y_AXIS || axis == COLOR_AXIS)
 	axis_array[axis].label.rotate = TEXT_VERTICAL;
 }
 
