@@ -1,5 +1,5 @@
 #ifndef lint
-static char *RCSid() { return RCSid("$Id: util3d.c,v 1.27.2.1 2007/03/22 04:37:11 sfeam Exp $"); }
+static char *RCSid() { return RCSid("$Id: util3d.c,v 1.27.2.3 2008/03/12 03:09:18 sfeam Exp $"); }
 #endif
 
 /* GNUPLOT - util3d.c */
@@ -950,8 +950,15 @@ map3d_xy(
     if (w == 0)
 	w = 1e-5;
 
-    *xt = (unsigned int) ((res[0] * xscaler / w) + xmiddle);
-    *yt = (unsigned int) ((res[1] * yscaler / w) + ymiddle);
+    if (lmargin.scalex == screen || rmargin.scalex == screen)
+	*xt = res[0] * xscaler/w + xmiddle;
+    else
+	*xt = (unsigned int) ((res[0] * xscaler / w) + xmiddle);
+
+    if (tmargin.scalex == screen || bmargin.scalex == screen)
+	*yt = res[1] * yscaler/w + ymiddle;
+    else
+	*yt = (unsigned int) ((res[1] * yscaler / w) + ymiddle);
 }
 
 /* Function to map from user 3D space to normalized 'camera' view
@@ -981,8 +988,15 @@ map3d_xy_double(
     if (w == 0)
 	w = 1e-5;
 
-    *xt = ((res[0] * xscaler / w) + xmiddle);
-    *yt = ((res[1] * yscaler / w) + ymiddle);
+    if (lmargin.scalex == screen || rmargin.scalex == screen)
+	*xt = res[0] * xscaler + xmiddle;
+    else
+	*xt = (res[0] * xscaler / w) + xmiddle;
+
+    if (tmargin.scalex == screen || bmargin.scalex == screen)
+	*yt = res[1] * yscaler + ymiddle;
+    else
+	*yt = (res[1] * yscaler / w) + ymiddle;
 }
 
 #else /* REPLICATE_CODE_FOR_BACKWARD_COMPATIBLE_ROUNDING */
@@ -1066,6 +1080,7 @@ draw3d_line_unconditional(
     int linetype)
 {
     unsigned int x1, y1, x2, y2;
+    struct lp_style_type ls = *lp;
 
     /* HBB 20020312: v2 can be NULL, if this call is coming from
     draw_line_hidden. --> redirect to point drawing routine */
@@ -1076,17 +1091,20 @@ draw3d_line_unconditional(
 
     TERMCOORD(v1, x1, y1);
     TERMCOORD(v2, x2, y2);
-    term_apply_lp_properties(lp);
 
-    /* FIXME - This whole set of tests may now be unnecessary */
-    if (lp->use_palette) {
-	double z =  (v1->real_z + v2->real_z) * 0.5;
-	if (lp->pm3d_color.type != TC_RGB)
-	    set_color(cb2gray(z2cb(z)));
-    } else
-	if (linetype != lp->l_type)
-	    (term->linetype)(linetype);
+    /* User-specified line styles */
+    if (prefer_line_styles && linetype >= 0)
+	lp_use_properties(&ls, linetype+1, FALSE);
 
+    /* The usual case of auto-generated line types */
+    else
+	ls.l_type = linetype;
+
+    /* Color by Z value */
+    if (ls.pm3d_color.type == TC_Z)
+	    ls.pm3d_color.value = (v1->real_z + v2->real_z) * 0.5;
+
+    term_apply_lp_properties(&ls);
     draw_clip_line(x1,y1,x2,y2);
 }
 
