@@ -1,5 +1,5 @@
 #ifndef lint
-static char *RCSid() { return RCSid("$Id: wgraph.c,v 1.37 2004/04/13 17:24:06 broeker Exp $"); }
+static char *RCSid() { return RCSid("$Id: wgraph.c,v 1.52.2.5 2008/09/09 16:42:50 sfeam Exp $"); }
 #endif
 
 /* GNUPLOT - win/wgraph.c */
@@ -35,7 +35,7 @@ static char *RCSid() { return RCSid("$Id: wgraph.c,v 1.37 2004/04/13 17:24:06 br
 
 /*
  * AUTHORS
- * 
+ *
  *   Maurice Castro
  *   Russell Lang
  */
@@ -59,10 +59,8 @@ static char *RCSid() { return RCSid("$Id: wgraph.c,v 1.37 2004/04/13 17:24:06 br
 # include "gpexecute.h"
 # include "mouse.h"
 #endif
-#ifdef PM3D
 # include "color.h"
-# include "getcolor.h"         
-#endif
+# include "getcolor.h"
 
 #ifdef USE_MOUSE
 /* Petr Mikulik, February 2001
@@ -79,6 +77,12 @@ static struct Ruler {
     int x, y;			/* ruler position */
 } ruler = {FALSE,0,0,};
 
+/* Status of the line from ruler to cursor */
+static struct RulerLineTo {
+    TBOOLEAN on;		/* ruler line active ? */
+    int x, y;			/* ruler line end position (previous cursor position) */
+} ruler_lineto = {FALSE,0,0,};
+
 /* Status of zoom box */
 static struct Zoombox {
     TBOOLEAN on;		/* set to TRUE during zooming */
@@ -92,11 +96,13 @@ HCURSOR hptrDefault, hptrCrossHair, hptrScaling, hptrRotating, hptrZooming, hptr
 /* Mouse support routines */
 static void	Wnd_exec_event(LPGW lpgw, LPARAM lparam, char type, int par1);
 static void	Wnd_refresh_zoombox(LPGW lpgw, LPARAM lParam);
+static void	Wnd_refresh_ruler_lineto(LPGW lpgw, LPARAM lParam);
 static void     GetMousePosViewport(LPGW lpgw, int *mx, int *my);
 static void	Draw_XOR_Text(LPGW lpgw, const char *text, size_t length, int x, int y);
 static void     DisplayStatusLine(LPGW lpgw);
 static void     UpdateStatusLine(LPGW lpgw, const char text[]);
 static void	DrawRuler(LPGW lpgw);
+static void	DrawRulerLineTo(LPGW lpgw);
 static void     DrawZoomBox(LPGW lpgw);
 static void	LoadCursors(LPGW lpgw);
 static void	DestroyCursors(LPGW lpgw);
@@ -133,7 +139,6 @@ int wginitstyle[WGDEFSTYLE] = {PS_SOLID, PS_DASH, PS_DOT, PS_DASHDOT, PS_DASHDOT
 #define GWOPMAX 4096
 
 
-#if USE_ULIG_FILLEDBOXES
 /* bitmaps for filled boxes (ULIG) */
 /* zeros represent the foreground color and ones represent the background color */
 /* FIXME HBB 20010916: *never* extern in a C source! */
@@ -141,7 +146,7 @@ int wginitstyle[WGDEFSTYLE] = {PS_SOLID, PS_DASH, PS_DOT, PS_DASHDOT, PS_DASHDOT
 /*  extern int fillpattern; */
 
 static unsigned char halftone_bitmaps[][16] ={
-  { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  
+  { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
     0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF },   /* no fill */
   { 0xEE, 0xEE, 0xBB, 0xBB, 0xEE, 0xEE, 0xBB, 0xBB,
     0xEE, 0xEE, 0xBB, 0xBB, 0xEE, 0xEE, 0xBB, 0xBB },   /* 25% pattern */
@@ -187,7 +192,6 @@ static BITMAP pattern_bitdata[pattern_num];
 static HBITMAP pattern_bitmap[pattern_num];
 
 static TBOOLEAN brushes_initialized = FALSE;
-#endif /* USE_ULIG_FILLEDBOXES */
 
 
 /* ================================== */
@@ -296,9 +300,19 @@ AddBlock(LPGW lpgw)
 void WDPROC
 GraphOp(LPGW lpgw, WORD op, WORD x, WORD y, LPCSTR str)
 {
-    struct GWOPBLK *this;
-    struct GWOP FAR *gwop;
-    char *npstr;
+    if (str)
+	GraphOpSize(lpgw, op, x, y, str, _fstrlen(str)+1);
+    else
+	GraphOpSize(lpgw, op, x, y, NULL, 0);
+}
+
+
+void WDPROC
+GraphOpSize(LPGW lpgw, WORD op, WORD x, WORD y, LPCSTR str, DWORD size)
+{
+	struct GWOPBLK *this;
+	struct GWOP FAR *gwop;
+	char *npstr;
 
 	this = lpgw->gwopblk_tail;
 	if ( (this==NULL) || (this->used >= GWOPMAX) ) {
@@ -313,10 +327,10 @@ GraphOp(LPGW lpgw, WORD op, WORD x, WORD y, LPCSTR str)
 	gwop->y = y;
 	gwop->htext = 0;
 	if (str) {
-		gwop->htext = LocalAlloc(LHND, _fstrlen(str)+1);
+		gwop->htext = LocalAlloc(LHND, size);
 		npstr = LocalLock(gwop->htext);
 		if (gwop->htext && (npstr != (char *)NULL))
-			lstrcpy(npstr, str);
+			memcpy(npstr, str, size);
 		LocalUnlock(gwop->htext);
 	}
 	this->used++;
@@ -333,7 +347,7 @@ GraphInit(LPGW lpgw)
 {
 	HMENU sysmenu;
 	WNDCLASS wndclass;
-	char buf[80];
+	char buf[MAX_PATH];
 
 	if (!lpgw->hPrevInstance) {
 		wndclass.style = CS_HREDRAW | CS_VREDRAW;
@@ -358,9 +372,9 @@ GraphInit(LPGW lpgw)
 		NULL, NULL, lpgw->hInstance, lpgw);
 
 	lpgw->hPopMenu = CreatePopupMenu();
-	AppendMenu(lpgw->hPopMenu, MF_STRING | (lpgw->graphtotop ? MF_CHECKED : MF_UNCHECKED), 
+	AppendMenu(lpgw->hPopMenu, MF_STRING | (lpgw->graphtotop ? MF_CHECKED : MF_UNCHECKED),
 		M_GRAPH_TO_TOP, "Bring to &Top");
-	AppendMenu(lpgw->hPopMenu, MF_STRING | (lpgw->color ? MF_CHECKED : MF_UNCHECKED), 
+	AppendMenu(lpgw->hPopMenu, MF_STRING | (lpgw->color ? MF_CHECKED : MF_UNCHECKED),
 		M_COLOR, "C&olor");
 	AppendMenu(lpgw->hPopMenu, MF_STRING, M_COPY_CLIP, "&Copy to Clipboard");
 #if WINVER >= 0x030a
@@ -402,7 +416,7 @@ GraphClose(LPGW lpgw)
 	DestroyBlocks(lpgw);
 	lpgw->locked = FALSE;
 }
-	
+
 
 void WDPROC
 GraphStart(LPGW lpgw, double pointsize)
@@ -420,15 +434,15 @@ GraphStart(LPGW lpgw, double pointsize)
 #ifdef USE_MOUSE
 		if (mouse_setting.on) {
 			BringWindowToTop(lpgw->hWndGraph);
-			return; 
+			return;
 		}
-#endif /* USE_MOUSE */		
-		SetWindowPos(lpgw->hWndGraph, 
+#endif /* USE_MOUSE */
+		SetWindowPos(lpgw->hWndGraph,
 			     HWND_TOP, 0,0,0,0,
 			     SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);
 	}
 }
-		
+
 void WDPROC
 GraphEnd(LPGW lpgw)
 {
@@ -439,7 +453,7 @@ GraphEnd(LPGW lpgw)
 	lpgw->locked = FALSE;
 	UpdateWindow(lpgw->hWndGraph);
 #ifdef USE_MOUSE
-	gp_exec_event(GE_plotdone, 0, 0, 0, 0);	/* notify main program */
+	gp_exec_event(GE_plotdone, 0, 0, 0, 0, 0);	/* notify main program */
 #endif
 }
 
@@ -497,7 +511,7 @@ StorePen(LPGW lpgw, int i, COLORREF ref, int colorstyle, int monostyle)
 }
 
 /* Prepare pens and brushes (--> colors) for use by the driver. Pens are (now) created
- * on-the-fly (--> DeleteObject(SelectObject(...)) idiom), but the brushes are still 
+ * on-the-fly (--> DeleteObject(SelectObject(...)) idiom), but the brushes are still
  * all created statically, and kept until the window is closed */
 static void
 MakePens(LPGW lpgw, HDC hdc)
@@ -516,7 +530,6 @@ MakePens(LPGW lpgw, HDC hdc)
 			lpgw->colorbrush[i] = CreateSolidBrush(lpgw->colorpen[i].lopnColor);
 	}
 
-#if USE_ULIG_FILLEDBOXES
 	/* build pattern brushes for filled boxes (ULIG) */
 	if( ! brushes_initialized ) {
 		int i;
@@ -547,7 +560,6 @@ MakePens(LPGW lpgw, HDC hdc)
 
 		brushes_initialized = TRUE;
 	}
-#endif /* USE_ULIG_FILLEDBOXES */
 }
 
 /* Undo effect of MakePens(). To be called just before the window is closed. */
@@ -561,7 +573,6 @@ DestroyPens(LPGW lpgw)
 	for (i=0; i<WGNUMPENS+2; i++)
 		DeleteObject(lpgw->colorbrush[i]);
 
-#if USE_ULIG_FILLEDBOXES
 	/* delete brushes used for boxfilling (ULIG) */
 	if( brushes_initialized ) {
 		int i;
@@ -576,7 +587,6 @@ DestroyPens(LPGW lpgw)
 		}
 		brushes_initialized = FALSE;
 	}
-#endif /* USE_ULIG_FILLEDBOXES */
 }
 
 /* ================================== */
@@ -606,7 +616,6 @@ Wnd_GetTextSize(HDC hdc, LPCSTR str, size_t len, int *cx, int *cy)
 static void
 MakeFonts(LPGW lpgw, LPRECT lprect, HDC hdc)
 {
-	LOGFONT lf;
 	HFONT hfontold;
 	TEXTMETRIC tm;
 	int result;
@@ -614,28 +623,29 @@ MakeFonts(LPGW lpgw, LPRECT lprect, HDC hdc)
 	int cx, cy;
 
 	lpgw->rotate = FALSE;
-	_fmemset(&lf, 0, sizeof(LOGFONT));
-	_fstrncpy(lf.lfFaceName,lpgw->fontname,LF_FACESIZE);
-	lf.lfHeight = -MulDiv(lpgw->fontsize, GetDeviceCaps(hdc, LOGPIXELSY), 72);
-	lf.lfCharSet = DEFAULT_CHARSET;
+	_fmemset(&(lpgw->lf), 0, sizeof(LOGFONT));
+	_fstrncpy(lpgw->lf.lfFaceName,lpgw->fontname,LF_FACESIZE);
+	lpgw->lf.lfHeight = -MulDiv(lpgw->fontsize, GetDeviceCaps(hdc, LOGPIXELSY), 72);
+	lpgw->lf.lfCharSet = DEFAULT_CHARSET;
 	if ( (p = _fstrstr(lpgw->fontname," Italic")) != (LPSTR)NULL ) {
-		lf.lfFaceName[ (unsigned int)(p-lpgw->fontname) ] = '\0';
-		lf.lfItalic = TRUE;
+		lpgw->lf.lfFaceName[ (unsigned int)(p-lpgw->fontname) ] = '\0';
+		lpgw->lf.lfItalic = TRUE;
 	}
 	if ( (p = _fstrstr(lpgw->fontname," Bold")) != (LPSTR)NULL ) {
-		lf.lfFaceName[ (unsigned int)(p-lpgw->fontname) ] = '\0';
-		lf.lfWeight = FW_BOLD;
+		lpgw->lf.lfFaceName[ (unsigned int)(p-lpgw->fontname) ] = '\0';
+		lpgw->lf.lfWeight = FW_BOLD;
 	}
 
 	if (lpgw->hfonth == 0) {
-		lpgw->hfonth = CreateFontIndirect((LOGFONT FAR *)&lf);
+		lpgw->hfonth = CreateFontIndirect((LOGFONT FAR *)&(lpgw->lf));
 	}
 
-	if (lpgw->hfontv == 0) {
-		lf.lfEscapement = 900;
-		lf.lfOrientation = 900;
-		lpgw->hfontv = CreateFontIndirect((LOGFONT FAR *)&lf);
-	}
+	/* we do need a 90 degree font */
+	if (lpgw->hfontv) 
+		DeleteObject(lpgw->hfontv);
+	lpgw->lf.lfEscapement = 900;
+	lpgw->lf.lfOrientation = 900;
+	lpgw->hfontv = CreateFontIndirect((LOGFONT FAR *)&(lpgw->lf));
 
 	/* save text size */
 	hfontold = SelectObject(hdc, lpgw->hfonth);
@@ -680,24 +690,28 @@ DestroyFonts(LPGW lpgw)
 static void
 SetFont(LPGW lpgw, HDC hdc)
 {
-	if (lpgw->rotate && lpgw->angle) {
-		if (lpgw->hfontv)
-			SelectObject(hdc, lpgw->hfontv);
-	} else {
-		if (lpgw->hfonth)
-			SelectObject(hdc, lpgw->hfonth);
-	}
-	return;
+    if (lpgw->rotate && lpgw->angle) {
+	if (lpgw->hfontv)
+	    DeleteObject(lpgw->hfontv);
+	lpgw->lf.lfEscapement = lpgw->lf.lfOrientation  = lpgw->angle * 10;
+	lpgw->hfontv = CreateFontIndirect((LOGFONT FAR *)&(lpgw->lf));
+       if (lpgw->hfontv)
+	    SelectObject(hdc, lpgw->hfontv);
+    } else {
+	if (lpgw->hfonth)
+	    SelectObject(hdc, lpgw->hfonth);
+    }
+    return;
 }
 
 static void
-SelFont(LPGW lpgw) 
+SelFont(LPGW lpgw)
 {
 #if WINVER >= 0x030a
 	LOGFONT lf;
 	CHOOSEFONT cf;
 	HDC hdc;
-	char lpszStyle[LF_FACESIZE]; 
+	char lpszStyle[LF_FACESIZE];
 	char FAR *p;
 
 	/* Set all structure fields to zero. */
@@ -729,6 +743,9 @@ SelFont(LPGW lpgw)
 			lstrcat(lpgw->fontname," Bold");
 		if (cf.nFontType & ITALIC_FONTTYPE)
 			lstrcat(lpgw->fontname," Italic");
+		/* set current font as default font */
+		strcpy(lpgw->deffontname,lpgw->fontname);
+		lpgw->deffontsize = lpgw->fontsize;
 		SendMessage(lpgw->hWndGraph,WM_COMMAND,M_REBUILDTOOLS,0L);
 	}
 #endif
@@ -790,19 +807,17 @@ drawgraph(LPGW lpgw, HDC hdc, LPRECT rect)
     unsigned int ngwop=0;
     BOOL isColor;
     double line_width = 1.0;
-#if USE_ULIG_FILLEDBOXES
     unsigned int fillstyle = 0.0;
     int idx;
-#endif /*  USE_ULIG_FILLEDBOXES */
 
-    if (lpgw->locked) 
+    if (lpgw->locked)
 	return;
 
     /* HBB 20010218: the GDI status query functions don't work on Metafile
      * handles, so can't know whether the screen is actually showing
      * color or not, if drawgraph() is being called from CopyClip().
      * Solve by defaulting isColor to 1 if hdc is a metafile. */
-    isColor = (((GetDeviceCaps(hdc, PLANES) * GetDeviceCaps(hdc,BITSPIXEL)) 
+    isColor = (((GetDeviceCaps(hdc, PLANES) * GetDeviceCaps(hdc,BITSPIXEL))
 		> 2)
 	       || (GetDeviceCaps(hdc, TECHNOLOGY) == DT_METAFILE));
 
@@ -818,17 +833,17 @@ drawgraph(LPGW lpgw, HDC hdc, LPRECT rect)
     rt = rect->top;
     rb = rect->bottom;
 
-    htic = lpgw->org_pointsize*MulDiv(lpgw->htic, rr-rl, lpgw->xmax) + 1;
-    vtic = lpgw->org_pointsize*MulDiv(lpgw->vtic, rb-rt, lpgw->ymax) + 1;
+    htic = lpgw->org_pointsize*MulDiv(lpgw->htic, rr - rl, lpgw->xmax) + 1;
+    vtic = lpgw->org_pointsize*MulDiv(lpgw->vtic, rb - rt, lpgw->ymax) + 1;
 
     lpgw->angle = 0;
     SetFont(lpgw, hdc);
-    SetTextAlign(hdc, TA_LEFT|TA_BOTTOM);
-    vshift = MulDiv(lpgw->vchar, rb-rt, lpgw->ymax)/2;
-    /* HBB 980630: new variable for moving rotated text to the correct
-     * position: */
-    hshift = MulDiv(lpgw->vchar, rr-rl, lpgw->xmax)/2;
-
+    SetTextAlign(hdc, TA_LEFT|TA_BASELINE);
+ 
+    /* calculate text shifting for horizontal text */
+    hshift = 0;
+    vshift = MulDiv(lpgw->vchar, rb - rt, lpgw->ymax)/2;
+  
     pen = 2;
     SelectObject(hdc, lpgw->hapen);
     SelectObject(hdc, lpgw->colorbrush[pen]);
@@ -849,8 +864,11 @@ drawgraph(LPGW lpgw, HDC hdc, LPRECT rect)
 	xdash = MulDiv(curptr->x, rr-rl-1, lpgw->xmax) + rl;
 	ydash = MulDiv(curptr->y, rt-rb+1, lpgw->ymax) + rb - 1;
 	if ((lastop==W_vect) && (curptr->op!=W_vect)) {
-	    if (polyi >= 2)
+	    if (polyi >= 2) {
 		Polyline(hdc, ppt, polyi);
+		MoveTo(hdc, ppt[0].x, ppt[0].y);
+	    } else if (polyi == 1)
+		LineTo(hdc, ppt[0].x, ppt[0].y);
 	    polyi = 0;
 	}
 	switch (curptr->op) {
@@ -874,34 +892,47 @@ drawgraph(LPGW lpgw, HDC hdc, LPRECT rect)
 	    break;
 	case W_line_type:
 	    {
+		LOGBRUSH lb;
+		LOGPEN cur_penstruct;
+
 		short cur_pen = ((curptr->x < (WORD)(-2))
 				 ? (curptr->x % WGNUMPENS) + 2
 				 : curptr->x + 2);
-		LOGPEN cur_penstruct = (lpgw->color && isColor) ?
-		    lpgw->colorpen[cur_pen] : lpgw->monopen[cur_pen];
+		/* set color only when second parameter to W_line_type equals 1 */
+		if (curptr->y != 1)
+		    pen = cur_pen;
+		cur_penstruct = (lpgw->color && isColor) ?
+		    lpgw->colorpen[pen] : lpgw->monopen[pen];
+		cur_penstruct.lopnColor = ((lpgw->color && isColor) ?
+		    lpgw->colorpen[cur_pen] : lpgw->monopen[cur_pen]).lopnColor;
 
 		if (line_width != 1)
 		    cur_penstruct.lopnWidth.x *= line_width;
-		lpgw->hapen = CreatePenIndirect((LOGPEN FAR *) &cur_penstruct);
+	
+		/* use ExtCreatePen instead of CreatePen/CreatePenIndirect to support
+		 * dashed lines if line_width > 1 */
+		lb.lbStyle = BS_SOLID;
+		lb.lbColor = cur_penstruct.lopnColor;
+		lpgw->hapen = ExtCreatePen(
+		        (line_width==1 ? PS_COSMETIC : PS_GEOMETRIC) | cur_penstruct.lopnStyle | PS_ENDCAP_FLAT | PS_JOIN_BEVEL, 
+			cur_penstruct.lopnWidth.x, &lb, 0, 0);
 		DeleteObject(SelectObject(hdc, lpgw->hapen));
 
-		pen = cur_pen;
-		SelectObject(hdc, lpgw->colorbrush[pen]);
+		SelectObject(hdc, lpgw->colorbrush[cur_pen]);
 		/* PM 7.7.2002: support color text */
-		SetTextColor(hdc, lpgw->colorpen[pen].lopnColor);
+		SetTextColor(hdc, cur_penstruct.lopnColor);
 	    }
 	break;
-	    
+
 	case W_put_text:
 	    {
 		char *str;
 		str = LocalLock(curptr->htext);
 		if (str) {
-		    /* HBB 980630: shift differently for rotated text: */
-		    if (lpgw->angle)
-			xdash += hshift;
-		    else
-			ydash += vshift;
+		    /* shift correctly for rotated text */
+		    xdash += hshift;
+		    ydash += vshift;
+
 		    SetBkMode(hdc,TRANSPARENT);
 		    TextOut(hdc,xdash,ydash,str,lstrlen(str));
 		    SetBkMode(hdc,OPAQUE);
@@ -910,7 +941,6 @@ drawgraph(LPGW lpgw, HDC hdc, LPRECT rect)
 	    }
 	break;
 
-#if USE_ULIG_FILLEDBOXES
 	case W_fillstyle:
 	    /* HBB 20010916: new entry, needed to squeeze the many
 	     * parameters of a filled box call through the bottlneck
@@ -918,19 +948,17 @@ drawgraph(LPGW lpgw, HDC hdc, LPRECT rect)
 	     * struct GWOP, respectively. */
 	    fillstyle = curptr->x;
 	    break;
-#endif /* USE_ULIG_FILLEDBOXES */
 
 	case W_boxfill:   /* ULIG */
 
 	    assert (polyi == 1);
-		
-#if USE_ULIG_FILLEDBOXES
+
 	    /* NOTE: the x and y passed with this call are the width and
 	     * height of the box, actually. The left corner was stored into
 	     * ppt[0] by a preceding W_move, and the style was memorized
 	     * by a W_fillstyle call. */
 	    switch(fillstyle & 0x0f) {
-		case FS_SOLID: 
+		case FS_SOLID:
 		    /* style == 1 --> use halftone fill pattern
 		     * according to filldensity. Density is from
 		     * 0..100 percent: */
@@ -941,11 +969,11 @@ drawgraph(LPGW lpgw, HDC hdc, LPRECT rect)
 			idx = halftone_num - 1;
 		    SelectObject(hdc, halftone_brush[idx]);
 		    break;
-		case FS_PATTERN: 
+		case FS_PATTERN:
 		    /* style == 2 --> use fill pattern according to
                      * fillpattern. Pattern number is enumerated */
-		    idx = fillstyle >> 4;  
-		    if (idx < 0) 
+		    idx = fillstyle >> 4;
+		    if (idx < 0)
 			idx = 0;
 		    if (idx > pattern_num - 1)
 			idx = 0;
@@ -957,43 +985,47 @@ drawgraph(LPGW lpgw, HDC hdc, LPRECT rect)
 		    SelectObject(hdc, halftone_brush[0]);
 	    }
 	    /* needs to be fixed for monochrome devices */
+	    /* FIXME: probably should keep track of text color */
 	    SetTextColor(hdc, lpgw->colorpen[pen].lopnColor);
 	    xdash -= rl;
 	    ydash -= rb - 1;
 	    PatBlt(hdc, ppt[0].x, ppt[0].y, xdash, ydash, PATCOPY);
-#else /* ! USE_ULIG_FILLEDBOXES */
-	    {
-		RECT rect;
-
-		rect.left = ppt[0].x;
-		rect.bottom = ppt[0].y;
-		rect.right = ppt[0].x + xdash;
-		rect.top = ppt[0].y - ydash;
-
-		SetBkColor(hdc,lpgw->background);
-		FillRect(hdc, &rect, lpgw->hbrush);
-	    }
-#endif /* USE_ULIG_FILLEDBOXES */
 	    polyi = 0;
 	    break;
-	case W_text_angle:
-	    lpgw->angle = curptr->x;
-	    SetFont(lpgw,hdc);
+  	case W_text_angle:
+ 	    if (lpgw->angle != (short int)curptr->x) {
+ 		lpgw->angle = (short int)curptr->x;
+ 		/* correctly calculate shifting of rotated text */
+ 		hshift = sin(M_PI/180. * lpgw->angle) * MulDiv(lpgw->vchar, rr-rl, lpgw->xmax) / 2;
+ 		vshift = cos(M_PI/180. * lpgw->angle) * MulDiv(lpgw->vchar, rb-rt, lpgw->ymax) / 2;
+ 		SetFont(lpgw, hdc);
+ 	    }
 	    break;
 	case W_justify:
 	    switch (curptr->x)
 		{
 		case LEFT:
-		    SetTextAlign(hdc, TA_LEFT|TA_BOTTOM);
+		    SetTextAlign(hdc, TA_LEFT|TA_BASELINE);
 		    break;
 		case RIGHT:
-		    SetTextAlign(hdc, TA_RIGHT|TA_BOTTOM);
+		    SetTextAlign(hdc, TA_RIGHT|TA_BASELINE);
 		    break;
 		case CENTRE:
-		    SetTextAlign(hdc, TA_CENTER|TA_BOTTOM);
+		    SetTextAlign(hdc, TA_CENTER|TA_BASELINE);
 		    break;
 		}
 	    break;
+	case W_font: {
+		char *font;
+
+		font = LocalLock(curptr->htext);
+		if (font) {
+		    GraphChangeFont(lpgw, font, curptr->x, hdc, *rect);
+	            SetFont(lpgw, hdc);
+		}
+		LocalUnlock(curptr->htext);
+	    }
+	    break;
 	case W_pointsize:
 	    if (curptr->x != 0) {
 		double pointsize = curptr->x / 100.0;
@@ -1019,36 +1051,45 @@ drawgraph(LPGW lpgw, HDC hdc, LPRECT rect)
 	     * state */
 	    line_width = curptr->x == 100 ? 1 : (curptr->x / 100.0);
 	    break;
-#ifdef PM3D /* HBB 20000813: implemented PM3D for Windows terminal */
 	case W_pm3d_setcolor:
 	    {
-		double level = curptr->x / 256.0;
-		rgb255_color rgb255;
 		static HBRUSH last_pm3d_brush = NULL;
 		HBRUSH this_brush;
 		COLORREF c;
+		LOGPEN cur_penstruct;
+		LOGBRUSH lb;
 
-		if (sm_palette.use_maxcolors > 0 && sm_palette.use_maxcolors < 256) {
-		    /* number of palette colours is reduced */
-		    if (curptr->x >=255)
-			level = 1;
-		    else {
-			level = (int)(level * sm_palette.colors);
-			level = (level >= sm_palette.colors-1) ? 1 : level / sm_palette.colors;
-		    }
+		/* distinguish gray values and RGB colors */
+		if (curptr->y == 0) {
+		    rgb255_color rgb255;
+		    rgb255maxcolors_from_gray( curptr->x / 256.0, &rgb255 );
+		    c = RGB(rgb255.r, rgb255.g, rgb255.b);
+		}
+		else {
+		    c = RGB(curptr->y & 0xff, (curptr->x >> 8) & 0xff, curptr->x & 0xff);
 		}
 
-		rgb255_from_gray( level, &rgb255 );
-		c = RGB(rgb255.r, rgb255.g, rgb255.b);
+		/* FIXME: always a _solid_ brush?? */
 		this_brush = CreateSolidBrush(c);
 		SelectObject(hdc, this_brush);
 		if (last_pm3d_brush != NULL)
 		    DeleteObject(last_pm3d_brush);
 		last_pm3d_brush = this_brush;
 		/* create new pen, too: */
-		DeleteObject(SelectObject(hdc, CreatePen(PS_SOLID, 1, c)));
+		cur_penstruct = (lpgw->color && isColor) ?
+		    lpgw->colorpen[pen] : lpgw->monopen[pen];	
+		if (line_width != 1)
+		    cur_penstruct.lopnWidth.x *= line_width;
+		lb.lbStyle = BS_SOLID;
+		lb.lbColor = c;
+		lpgw->hapen = ExtCreatePen(
+		    (line_width==1 ? PS_COSMETIC : PS_GEOMETRIC) | cur_penstruct.lopnStyle | PS_ENDCAP_FLAT | PS_JOIN_BEVEL, 
+		    cur_penstruct.lopnWidth.x, &lb, 0, 0);
+		DeleteObject(SelectObject(hdc, lpgw->hapen));
+		/* finally set text color */
+		SetTextColor(hdc, c);
 	    }
-	break;
+	    break;
 	case W_pm3d_filled_polygon_pt:
 	    {
 		/* a point of the polygon is coming */
@@ -1060,31 +1101,78 @@ drawgraph(LPGW lpgw, HDC hdc, LPRECT rect)
 		ppt[polyi].y = ydash;
 		polyi++;
 	    }
-	break;
+	    break;
 	case W_pm3d_filled_polygon_draw:
 	    {
 		/* end of point series --> draw polygon now */
 		Polygon(hdc, ppt, polyi);
 		polyi = 0;
 	    }
-	break;
-#endif /* PM3D */
+	    break;
+	case W_image:
+	    {
+		/* Due to the structure of gwop in total 5 entries are needed.
+		   These static variables help to collect all the needed information
+		*/
+		static int seq = 0;  /* sequence counter */
+		static POINT corners[4];
+
+		if (seq < 4) {
+		    /* The first four OPs contain the `corner` array */
+		    corners[seq].x = xdash;
+		    corners[seq].y = ydash;
+		} else {
+		    /* The last OP contains the image and it's size */
+		    BITMAPINFO bmi;
+		    char *image;
+		    unsigned int M, N;
+		    int rc;
+
+		    M = curptr->x;
+		    N = curptr->y;
+		    image = LocalLock(curptr->htext);
+		    if (image) {
+
+			/* rc = SetStretchBltMode(hdc, HALFTONE); */
+
+			memset(&bmi, 0, sizeof(bmi));
+			bmi.bmiHeader.biSize = sizeof(bmi.bmiHeader);
+			bmi.bmiHeader.biWidth = M;
+			bmi.bmiHeader.biHeight = N;
+			bmi.bmiHeader.biPlanes = 1;
+			bmi.bmiHeader.biBitCount = 24;
+    			bmi.bmiHeader.biCompression = BI_RGB;
+			bmi.bmiHeader.biClrUsed = 0;
+
+			rc = StretchDIBits(hdc, 
+			    corners[0].x, corners[0].y, 
+			    corners[1].x - corners[0].x, corners[1].y - corners[0].y,
+			    0, 0,
+			    M, N,
+			    image, &bmi,
+			    DIB_RGB_COLORS, SRCCOPY );
+		    }
+		    LocalUnlock(curptr->htext);
+		}
+		seq = (seq + 1) % 5;
+	    }
+	    break;
 	case W_dot:
 	    dot(hdc, xdash, ydash);
 	    break;
-	case W_plus: /* do plus */ 
+	case W_plus: /* do plus */
 	    MoveTo(hdc,xdash-htic,ydash);
 	    LineTo(hdc,xdash+htic+1,ydash);
 	    MoveTo(hdc,xdash,ydash-vtic);
 	    LineTo(hdc,xdash,ydash+vtic+1);
 	    break;
-	case W_cross: /* do X */ 
+	case W_cross: /* do X */
 	    MoveTo(hdc,xdash-htic,ydash-vtic);
 	    LineTo(hdc,xdash+htic+1,ydash+vtic+1);
 	    MoveTo(hdc,xdash-htic,ydash+vtic);
 	    LineTo(hdc,xdash+htic+1,ydash-vtic-1);
 	    break;
-	case W_star: /* do star */ 
+	case W_star: /* do star */
 	    MoveTo(hdc,xdash-htic,ydash);
 	    LineTo(hdc,xdash+htic+1,ydash);
 	    MoveTo(hdc,xdash,ydash-vtic);
@@ -1094,12 +1182,12 @@ drawgraph(LPGW lpgw, HDC hdc, LPRECT rect)
 	    MoveTo(hdc,xdash-htic,ydash+vtic);
 	    LineTo(hdc,xdash+htic+1,ydash-vtic-1);
 	    break;
-	case W_circle: /* do open circle */ 
+	case W_circle: /* do open circle */
 	    Arc(hdc, xdash-htic, ydash-vtic, xdash+htic+1, ydash+vtic+1,
 		xdash, ydash+vtic+1, xdash, ydash+vtic+1);
 	    dot(hdc, xdash, ydash);
 	    break;
-	case W_fcircle: /* do filled circle */ 
+	case W_fcircle: /* do filled circle */
 	    Ellipse(hdc, xdash-htic, ydash-vtic,
 		    xdash+htic+1, ydash+vtic+1);
 	    break;
@@ -1107,7 +1195,7 @@ drawgraph(LPGW lpgw, HDC hdc, LPRECT rect)
 	    {
 		POINT p[6];
 		int i;
-		int shape;
+		int shape = 0;
 		int filled = 0;
 		static float pointshapes[5][10] = {
 		    {-1, -1, +1, -1, +1, +1, -1, +1, 0, 0}, /* box */
@@ -1137,15 +1225,27 @@ drawgraph(LPGW lpgw, HDC hdc, LPRECT rect)
 		    shape = 4;
 		    break;
 		case W_fbox:
+		    shape = 0;
+		    filled = 1;
+		    break;
 		case W_fdiamond:
+		    shape = 1;
+		    filled = 1;
+		    break;
 		case W_fitriangle:
+		    shape = 2;
+		    filled = 1;
+		    break;
 		case W_ftriangle:
+		    shape = 3;
+		    filled = 1;
+		    break;
 		case W_fpentagon:
-		    shape = curptr->op-W_fbox;
+		    shape = 4;
 		    filled = 1;
 		    break;
 		}
-						
+
 		for (i = 0; i < 5; ++i) {
 		    if (pointshapes[shape][i * 2 + 1] == 0
 			&& pointshapes[shape][i * 2] == 0)
@@ -1321,7 +1421,7 @@ CopyPrint(LPGW lpgw)
 	PRINTDLG pd;
 	HWND hwnd;
 	RECT rect;
-	PRINT pr;
+	GP_PRINT pr;
 
 	hwnd = lpgw->hWndGraph;
 
@@ -1342,16 +1442,16 @@ CopyPrint(LPGW lpgw)
 	}
 
 	pr.hdcPrn = printer;
-	SetWindowLong(hwnd, 4, (LONG)((LPPRINT)&pr));
+	SetWindowLong(hwnd, 4, (LONG)((GP_LPPRINT)&pr));
 #ifdef WIN32
-	PrintRegister((LPPRINT)&pr);
+	PrintRegister((GP_LPPRINT)&pr);
 #endif
 
 	EnableWindow(hwnd,FALSE);
 	pr.bUserAbort = FALSE;
 #ifdef WIN32
 	pr.hDlgPrint = CreateDialogParam(hdllInstance,"CancelDlgBox",hwnd,PrintDlgProc,(LPARAM)lpgw->Title);
-	SetAbortProc(printer,PrintAbortProc);  
+	SetAbortProc(printer,PrintAbortProc);
 
 	memset(&docInfo, 0, sizeof(DOCINFO));
 	docInfo.cbSize = sizeof(DOCINFO);
@@ -1367,7 +1467,7 @@ CopyPrint(LPGW lpgw)
 	lpfnAbortProc = (DLGPROC)MakeProcInstance((FARPROC)PrintAbortProc, hdllInstance);
 #  endif /* __DLL__ */
 	pr.hDlgPrint = CreateDialogParam(hdllInstance,"CancelDlgBox",hwnd,lpfnPrintDlgProc,(LPARAM)lpgw->Title);
-	Escape(printer,SETABORTPROC,0,(LPSTR)lpfnAbortProc,NULL);  
+	Escape(printer,SETABORTPROC,0,(LPSTR)lpfnAbortProc,NULL);
 	if (Escape(printer, STARTDOC, lstrlen(lpgw->Title),lpgw->Title, NULL) > 0) {
 #endif
 		SetMapMode(printer, MM_TEXT);
@@ -1401,7 +1501,7 @@ CopyPrint(LPGW lpgw)
 	DeleteDC(printer);
 	SetWindowLong(hwnd, 4, (LONG)(0L));
 #ifdef WIN32
-	PrintUnregister((LPPRINT)&pr);
+	PrintUnregister((GP_LPPRINT)&pr);
 #endif /* WIN32 */
 	/* make certain that the screen pen set is restored */
 	SendMessage(lpgw->hWndGraph,WM_COMMAND,M_REBUILDTOOLS,0L);
@@ -1432,7 +1532,7 @@ WriteGraphIni(LPGW lpgw)
 	WritePrivateProfileString(section, "GraphOrigin", profile, file);
 	wsprintf(profile, "%d %d", rect.right-rect.left, rect.bottom-rect.top);
 	WritePrivateProfileString(section, "GraphSize", profile, file);
-	wsprintf(profile, "%s,%d", lpgw->fontname, lpgw->fontsize);
+	wsprintf(profile, "%s,%d", lpgw->deffontname, lpgw->deffontsize);
 	WritePrivateProfileString(section, "GraphFont", profile, file);
 	wsprintf(profile, "%d", lpgw->color);
 	WritePrivateProfileString(section, "GraphColor", profile, file);
@@ -1454,7 +1554,7 @@ WriteGraphIni(LPGW lpgw)
 		pm = &lpgw->monopen[i];
 		wsprintf(profile, "%d %d %d %d %d",GetRValue(pc->lopnColor),
 			GetGValue(pc->lopnColor), GetBValue(pc->lopnColor),
-			(pc->lopnWidth.x != 1) ? -pc->lopnWidth.x : pc->lopnStyle, 
+			(pc->lopnWidth.x != 1) ? -pc->lopnWidth.x : pc->lopnStyle,
 			(pm->lopnWidth.x != 1) ? -pm->lopnWidth.x : pm->lopnStyle);
 		WritePrivateProfileString(section, entry, profile, file);
 	}
@@ -1509,6 +1609,9 @@ ReadGraphIni(LPGW lpgw)
 			else
 				_fstrcpy(lpgw->fontname,WINFONT);
 		}
+		/* set current font as default font */
+		_fstrcpy(lpgw->deffontname, lpgw->fontname);
+		lpgw->deffontsize = lpgw->fontsize;
 	}
 
 	if (bOKINI)
@@ -1583,7 +1686,7 @@ typedef struct tagLS {
 	LOGPEN	monopen[WGNUMPENS+2];	/* logical mono pens */
 } LS;
 typedef LS FAR*  LPLS;
-	
+
 
 static COLORREF
 GetColor(HWND hwnd, COLORREF ref)
@@ -1649,41 +1752,41 @@ LineStyleDlgProc(HWND hdlg, UINT wmsg, WPARAM wparam, LPARAM lparam)
 					_fstrcpy(buf,"Axis");
 				else
 			 		wsprintf(buf,"Line%d",i-1);
-				SendDlgItemMessage(hdlg, LS_LINENUM, LB_ADDSTRING, 0, 
+				SendDlgItemMessage(hdlg, LS_LINENUM, LB_ADDSTRING, 0,
 					(LPARAM)((LPSTR)buf));
 			}
 			SendDlgItemMessage(hdlg, LS_LINENUM, LB_SETCURSEL, pen, 0L);
 
-			SendDlgItemMessage(hdlg, LS_MONOSTYLE, CB_ADDSTRING, 0, 
+			SendDlgItemMessage(hdlg, LS_MONOSTYLE, CB_ADDSTRING, 0,
 				(LPARAM)((LPSTR)"Solid"));
-			SendDlgItemMessage(hdlg, LS_MONOSTYLE, CB_ADDSTRING, 0, 
+			SendDlgItemMessage(hdlg, LS_MONOSTYLE, CB_ADDSTRING, 0,
 				(LPARAM)((LPSTR)"Dash"));
-			SendDlgItemMessage(hdlg, LS_MONOSTYLE, CB_ADDSTRING, 0, 
+			SendDlgItemMessage(hdlg, LS_MONOSTYLE, CB_ADDSTRING, 0,
 				(LPARAM)((LPSTR)"Dot"));
-			SendDlgItemMessage(hdlg, LS_MONOSTYLE, CB_ADDSTRING, 0, 
+			SendDlgItemMessage(hdlg, LS_MONOSTYLE, CB_ADDSTRING, 0,
 				(LPARAM)((LPSTR)"DashDot"));
-			SendDlgItemMessage(hdlg, LS_MONOSTYLE, CB_ADDSTRING, 0, 
+			SendDlgItemMessage(hdlg, LS_MONOSTYLE, CB_ADDSTRING, 0,
 				(LPARAM)((LPSTR)"DashDotDot"));
 
 			plpm = &lpls->monopen[pen];
-			SendDlgItemMessage(hdlg, LS_MONOSTYLE, CB_SETCURSEL, 
+			SendDlgItemMessage(hdlg, LS_MONOSTYLE, CB_SETCURSEL,
 				plpm->lopnStyle, 0L);
 			wsprintf(buf,"%d",plpm->lopnWidth.x);
 			SetDlgItemText(hdlg, LS_MONOWIDTH, buf);
 
-			SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_ADDSTRING, 0, 
+			SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_ADDSTRING, 0,
 				(LPARAM)((LPSTR)"Solid"));
-			SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_ADDSTRING, 0, 
+			SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_ADDSTRING, 0,
 				(LPARAM)((LPSTR)"Dash"));
-			SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_ADDSTRING, 0, 
+			SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_ADDSTRING, 0,
 				(LPARAM)((LPSTR)"Dot"));
-			SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_ADDSTRING, 0, 
+			SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_ADDSTRING, 0,
 				(LPARAM)((LPSTR)"DashDot"));
-			SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_ADDSTRING, 0, 
+			SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_ADDSTRING, 0,
 				(LPARAM)((LPSTR)"DashDotDot"));
 
 			plpc = &lpls->colorpen[pen];
-			SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_SETCURSEL, 
+			SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_SETCURSEL,
 				plpc->lopnStyle, 0L);
 			wsprintf(buf,"%d",plpc->lopnWidth.x);
 			SetDlgItemText(hdlg, LS_COLORWIDTH, buf);
@@ -1697,16 +1800,16 @@ LineStyleDlgProc(HWND hdlg, UINT wmsg, WPARAM wparam, LPARAM lparam)
 				case LS_LINENUM:
 					wsprintf(buf,"%d",plpm->lopnWidth.x);
 					SetDlgItemText(hdlg, LS_MONOWIDTH, buf);
-					SendDlgItemMessage(hdlg, LS_MONOSTYLE, CB_SETCURSEL, 
+					SendDlgItemMessage(hdlg, LS_MONOSTYLE, CB_SETCURSEL,
 						plpm->lopnStyle, 0L);
 					wsprintf(buf,"%d",plpc->lopnWidth.x);
 					SetDlgItemText(hdlg, LS_COLORWIDTH, buf);
-					SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_SETCURSEL, 
+					SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_SETCURSEL,
 						plpc->lopnStyle, 0L);
 					UpdateColorSample(hdlg);
 					return FALSE;
 				case LS_MONOSTYLE:
-					plpm->lopnStyle = 
+					plpm->lopnStyle =
 						(UINT)SendDlgItemMessage(hdlg, LS_MONOSTYLE, CB_GETCURSEL, 0, 0L);
 					if (plpm->lopnStyle != 0) {
 						plpm->lopnWidth.x = 1;
@@ -1719,7 +1822,7 @@ LineStyleDlgProc(HWND hdlg, UINT wmsg, WPARAM wparam, LPARAM lparam)
 					GetInt(buf, (LPINT)&plpm->lopnWidth.x);
 					if (plpm->lopnWidth.x != 1) {
 						plpm->lopnStyle = 0;
-						SendDlgItemMessage(hdlg, LS_MONOSTYLE, CB_SETCURSEL, 
+						SendDlgItemMessage(hdlg, LS_MONOSTYLE, CB_SETCURSEL,
 							plpm->lopnStyle, 0L);
 					}
 					return FALSE;
@@ -1728,7 +1831,7 @@ LineStyleDlgProc(HWND hdlg, UINT wmsg, WPARAM wparam, LPARAM lparam)
 					UpdateColorSample(hdlg);
 					return FALSE;
 				case LS_COLORSTYLE:
-					plpc->lopnStyle = 
+					plpc->lopnStyle =
 						(UINT)SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_GETCURSEL, 0, 0L);
 					if (plpc->lopnStyle != 0) {
 						plpc->lopnWidth.x = 1;
@@ -1741,7 +1844,7 @@ LineStyleDlgProc(HWND hdlg, UINT wmsg, WPARAM wparam, LPARAM lparam)
 					GetInt(buf, (LPINT)&plpc->lopnWidth.x);
 					if (plpc->lopnWidth.x != 1) {
 						plpc->lopnStyle = 0;
-						SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_SETCURSEL, 
+						SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_SETCURSEL,
 							plpc->lopnStyle, 0L);
 					}
 					return FALSE;
@@ -1780,7 +1883,7 @@ LineStyleDlgProc(HWND hdlg, UINT wmsg, WPARAM wparam, LPARAM lparam)
 						plpm->lopnStyle, 0L);
 					wsprintf(buf,"%d",plpc->lopnWidth.x);
 					SetDlgItemText(hdlg, LS_COLORWIDTH, buf);
-					SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_SETCURSEL, 
+					SendDlgItemMessage(hdlg, LS_COLORSTYLE, CB_SETCURSEL,
 						plpc->lopnStyle, 0L);
 					UpdateColorSample(hdlg);
 					return FALSE;
@@ -1819,7 +1922,7 @@ LineStyle(LPGW lpgw)
 #ifndef WIN32
 	DLGPROC lpfnLineStyleDlgProc;
 #endif
-	
+
 	SetWindowLong(lpgw->hWndGraph, 4, (LONG)((LPLS)&ls));
 	_fmemcpy(&ls.colorpen, &lpgw->colorpen, (WGNUMPENS + 2) * sizeof(LOGPEN));
 	_fmemcpy(&ls.monopen, &lpgw->monopen, (WGNUMPENS + 2) * sizeof(LOGPEN));
@@ -1862,7 +1965,7 @@ Wnd_exec_event(LPGW lpgw, LPARAM lparam, char type, int par1)
     unsigned long thisTimestamp = GetMessageTime();
 
     GetMousePosViewport(lpgw, &mx, &my);
-    gp_exec_event(type, mx, my, par1, thisTimestamp - lastTimestamp);
+    gp_exec_event(type, mx, my, par1, thisTimestamp - lastTimestamp, 0);
     lastTimestamp = thisTimestamp;
 }
 
@@ -1876,6 +1979,17 @@ Wnd_refresh_zoombox(LPGW lpgw, LPARAM lParam)
     zoombox.to.x = mx; zoombox.to.y = my;
     DrawZoomBox(lpgw); /*  draw new zoom box */
 }
+
+static void
+Wnd_refresh_ruler_lineto(LPGW lpgw, LPARAM lParam)
+{
+    int mx, my;
+
+    GetMousePosViewport(lpgw, &mx, &my);
+    DrawRulerLineTo(lpgw); /*  erase current line */
+    ruler_lineto.x = mx; ruler_lineto.y = my;
+    DrawRulerLineTo(lpgw); /*  draw new line box */
+}
 #endif /* USE_MOUSE */
 
 /* ================================== */
@@ -1909,6 +2023,9 @@ WndGraphProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 				if (zoombox.on) {
 					Wnd_refresh_zoombox(lpgw, lParam);
 				}
+				if (ruler.on && ruler_lineto.on) {
+					Wnd_refresh_ruler_lineto(lpgw, lParam);
+				}
 				/* track (show) mouse position -- send the event to gnuplot */
 				Wnd_exec_event(lpgw, lParam,  GE_motion, wParam);
 				return 0L; /* end of WM_MOUSEMOVE */
@@ -2254,6 +2371,7 @@ WndGraphProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 			 * doesn't know anything about has to be redrawn */
 			DrawRuler(lpgw);
 			DisplayStatusLine(lpgw);
+			DrawRulerLineTo(lpgw);
 			DrawZoomBox(lpgw);
 #endif
 			return 0;
@@ -2301,6 +2419,78 @@ WndGraphProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 	return DefWindowProc(hwnd, message, wParam, lParam);
 }
 
+
+void WDPROC
+GraphChangeFont(LPGW lpgw, LPCSTR font, int fontsize, HDC hdc, RECT rect)
+{
+    int newfontsize;
+    bool remakefonts = FALSE;
+
+    newfontsize = (fontsize != 0) ? fontsize : lpgw->deffontsize;
+    if (font != NULL) {
+	remakefonts = (strcmp(lpgw->fontname, font) != 0) || (newfontsize != lpgw->fontsize);
+    } else {
+	remakefonts = (strcmp(lpgw->fontname, lpgw->deffontname) != 0) || (newfontsize != lpgw->fontsize);
+    }
+
+    if (remakefonts) {
+        lpgw->fontsize = newfontsize;
+        strcpy(lpgw->fontname, (font) ? font : lpgw->deffontname);
+
+        DestroyFonts(lpgw);
+        MakeFonts(lpgw, &rect, hdc);
+    }
+}
+
+
+#if 0
+int WDPROC
+GraphGetFontScaling(LPGW lpgw, LPCSTR font, int fontsize)
+{
+    HDC hdc;
+    RECT rect;
+    HGDIOBJ hprevfont;
+    OUTLINETEXTMETRIC otm;
+    int shift = 35;
+
+    hdc = GetDC(lpgw->hWndGraph);
+    GetClientRect(lpgw->hWndGraph, &rect);
+    GraphChangeFont(lpgw, font, fontsize, hdc, rect);
+    hprevfont = SelectObject(hdc, lpgw->hfonth);
+    if (GetOutlineTextMetrics(hdc, sizeof(otm), &otm) != 0) {
+        shift = otm.otmptSuperscriptOffset.y - otm.otmptSubscriptOffset.y;
+        shift = MulDiv(shift, GetDeviceCaps(hdc, LOGPIXELSY), 2 * 72);
+        printf( "shift: %i (%i %i)\n", shift*8, otm.otmptSuperscriptOffset.y, otm.otmptSubscriptOffset.y);
+    }
+    SelectObject(hdc, hprevfont);
+
+    return shift * 8;
+}
+#endif
+
+
+unsigned int WDPROC
+GraphGetTextLength(LPGW lpgw, LPCSTR text, LPCSTR fontname, int fontsize)
+{
+    HDC hdc;
+    RECT rect;
+    SIZE size;
+    HGDIOBJ hprevfont;
+
+    hdc = GetDC(lpgw->hWndGraph);
+    GetClientRect(lpgw->hWndGraph, &rect);
+
+    GraphChangeFont(lpgw, fontname, fontsize, hdc, rect);
+
+    hprevfont = SelectObject(hdc, lpgw->hfonth);
+    GetTextExtentPoint(hdc, text, strlen(text), &size);
+    SelectObject(hdc, hprevfont);
+    
+    size.cx = MulDiv(size.cx + GetTextCharacterExtra(hdc), lpgw->xmax, rect.right-rect.left-1);
+    return size.cx;
+}
+
+
 #ifdef USE_MOUSE
 /* Implemented by Petr Mikulik, February 2001 --- the best Windows solutions
  * come from OS/2 :-))
@@ -2318,6 +2508,18 @@ void WDPROC
 Graph_set_cursor (LPGW lpgw, int c, int x, int y )
 {
 	switch (c) {
+	case -4: /* switch off line between ruler and mouse cursor */
+		DrawRulerLineTo(lpgw);
+		ruler_lineto.on = FALSE;
+		break;
+	case -3: /* switch on line between ruler and mouse cursor */
+		if (ruler.on && ruler_lineto.on)
+		    break;
+		ruler_lineto.x = x;
+		ruler_lineto.y = y;
+		ruler_lineto.on = TRUE;
+		DrawRulerLineTo(lpgw);
+		break;
 	case -2:
 		{ /* move mouse to the given point */
 			RECT rc;
@@ -2353,6 +2555,10 @@ Graph_set_cursor (LPGW lpgw, int c, int x, int y )
 		DrawZoomBox(lpgw);
 		zoombox.on = FALSE;
 	}
+	if (c>=0 && ruler_lineto.on) { /* erase ruler line */
+		DrawRulerLineTo(lpgw);
+		ruler_lineto.on = FALSE;
+	}
 }
 
 /* set_ruler(int x, int y) term API: x<0 switches ruler off. */
@@ -2360,6 +2566,7 @@ void WDPROC
 Graph_set_ruler (LPGW lpgw, int x, int y )
 {
 	DrawRuler(lpgw); /* remove previous drawing, if any */
+	DrawRulerLineTo(lpgw);
 	if (x < 0) {
 		ruler.on = FALSE;
 		return;
@@ -2367,6 +2574,7 @@ Graph_set_ruler (LPGW lpgw, int x, int y )
 	ruler.on = TRUE;
 	ruler.x = x; ruler.y = y;
 	DrawRuler(lpgw); /* draw ruler at new positions */
+	DrawRulerLineTo(lpgw);
 }
 
 /* put_tmptext(int i, char c[]) term API
@@ -2383,7 +2591,7 @@ Graph_put_tmptext (LPGW lpgw, int where, LPCSTR text )
 	case 0:
 		UpdateStatusLine(lpgw, text);
 		break;
-	case 1: 
+	case 1:
 		DrawZoomBox(lpgw);
 		if (zoombox.text1) {
 			free((char*)zoombox.text1);
@@ -2391,7 +2599,7 @@ Graph_put_tmptext (LPGW lpgw, int where, LPCSTR text )
 		zoombox.text1 = strdup(text);
 		DrawZoomBox(lpgw);
 		break;
-	case 2: 
+	case 2:
 		DrawZoomBox(lpgw);
 		if (zoombox.text2) {
 			free((char*)zoombox.text2);
@@ -2573,6 +2781,34 @@ DrawRuler (LPGW lpgw)
 	ReleaseDC(lpgw->hWndGraph, hdc);
 }
 
+/* Draw the ruler line to cursor position.
+ */
+static void
+DrawRulerLineTo (LPGW lpgw)
+{
+	HDC hdc;
+	int iOldRop;
+	RECT rc;
+	long rx, ry, rlx, rly;
+
+	if (!ruler.on || !ruler_lineto.on || ruler.x < 0 || ruler_lineto.x < 0)
+		return;
+
+	hdc = GetDC(lpgw->hWndGraph);
+	GetClientRect(lpgw->hWndGraph, &rc);
+
+	rx  = MulDiv(ruler.x, rc.right - rc.left, lpgw->xmax);
+	ry  = rc.bottom - MulDiv(ruler.y, rc.bottom - rc.top, lpgw->ymax);
+	rlx = MulDiv(ruler_lineto.x, rc.right - rc.left, lpgw->xmax);
+	rly = rc.bottom - MulDiv(ruler_lineto.y, rc.bottom - rc.top, lpgw->ymax);
+
+	iOldRop = SetROP2(hdc, R2_NOT);
+	MoveTo(hdc, rx, ry);
+	LineTo(hdc, rlx, rly);
+	SetROP2(hdc, iOldRop);
+	ReleaseDC(lpgw->hWndGraph, hdc);
+}
+
 /* Draw the zoom box.
  */
 static void
